# This file is part of ctrl_mpexec.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import click
from lsst.daf.butler.cli.utils import MWOptionDecorator, MWPath, split_commas, unwrap

butler_config_option = MWOptionDecorator(
    "-b", "--butler-config", help="Location of the gen3 butler/registry config file."
)


data_query_option = MWOptionDecorator(
    "-d", "--data-query", help="User data selection expression.", metavar="QUERY"
)


debug_option = MWOptionDecorator(
    "--debug", help="Enable debugging output using lsstDebug facility (imports debug.py).", is_flag=True
)

coverage_option = MWOptionDecorator("--coverage", help="Enable coverage output.", is_flag=True)

coverage_report_option = MWOptionDecorator(
    "--cov-report/--no-cov-report",
    help="If coverage is enabled, controls whether to produce an HTML coverage report.",
    default=True,
)

coverage_packages_option = MWOptionDecorator(
    "--cov-packages",
    help=unwrap(
        """Python packages to restrict coverage to.  If none are provided, runs coverage on all packages."""
    ),
    multiple=True,
    callback=split_commas,
)

delete_option = MWOptionDecorator(
    "--delete", callback=split_commas, help="Delete task with given label from pipeline.", multiple=True
)


pdb_option = MWOptionDecorator(
    "--pdb",
    help="Post-mortem debugger to launch for exceptions (defaults to pdb if unspecified; requires a tty).",
    is_flag=False,
    flag_value="pdb",
    default=None,
)


extend_run_option = MWOptionDecorator(
    "--extend-run",
    help=(
        "Instead of creating a new RUN collection, insert datasets into either the one given by "
        "--output-run (if provided) or the first child collection of --output (which must be of type RUN). "
        "This also enables --skip-existing option when building a graph. "
        "When executing a graph this option skips quanta with all existing outputs."
    ),
    is_flag=True,
)


graph_fixup_option = MWOptionDecorator(
    "--graph-fixup",
    help=unwrap(
        """Name of the class or factory method which makes an
                                                   instance used for execution graph fixup."""
    ),
)


init_only_option = MWOptionDecorator(
    "--init-only",
    help=unwrap(
        """Do not actually run; just register dataset types and/or
                                                 save init outputs. """
    ),
    is_flag=True,
)


input_option = MWOptionDecorator(
    "-i",
    "--input",
    callback=split_commas,
    default=list(),
    help=unwrap("""Comma-separated names of the input collection(s)."""),
    metavar="COLLECTION",
    multiple=True,
)

no_versions_option = MWOptionDecorator(
    "--no-versions", help="Do not save or check package versions.", is_flag=True
)


order_pipeline_option = MWOptionDecorator(
    "--order-pipeline",
    help=unwrap(
        """Order tasks in pipeline based on their data
                                          dependencies, ordering is performed as last step before saving or
                                          executing pipeline."""
    ),
    is_flag=True,
)


output_option = MWOptionDecorator(
    "-o",
    "--output",
    help=unwrap(
        """Name of the output CHAINED collection. This may either be an
                                              existing CHAINED collection to use as both input and output
                                              (incompatible with --input), or a new CHAINED collection created
                                              to include all inputs (requires --input). In both cases, the
                                              collection's children will start with an output RUN collection
                                              that directly holds all new datasets (see --output-run)."""
    ),
    metavar="COLL",
)


output_run_option = MWOptionDecorator(
    "--output-run",
    help=unwrap(
        """Name of the new output RUN collection. If not provided
                                                  then --output must be provided and a new RUN collection will
                                                  be created by appending a timestamp to the value passed with
                                                  --output. If this collection already exists then
                                                  --extend-run must be passed."""
    ),
    metavar="COLL",
)


pipeline_option = MWOptionDecorator(
    "-p",
    "--pipeline",
    help="Location of a pipeline definition file in YAML format.",
    type=MWPath(file_okay=True, dir_okay=False, readable=True),
)


pipeline_dot_option = MWOptionDecorator(
    "--pipeline-dot",
    help=unwrap(
        """"Location for storing GraphViz DOT representation of a
                                                    pipeline."""
    ),
    type=MWPath(writable=True, file_okay=True, dir_okay=False),
)


profile_option = MWOptionDecorator(
    "--profile", help="Dump cProfile statistics to file name.", type=MWPath(file_okay=True, dir_okay=False)
)


prune_replaced_option = MWOptionDecorator(
    "--prune-replaced",
    help=unwrap(
        """Delete the datasets in the collection replaced by
                                                      --replace-run, either just from the datastore
                                                      ('unstore') or by removing them and the RUN completely
                                                      ('purge'). Requires --replace-run."""
    ),
    type=click.Choice(choices=("unstore", "purge"), case_sensitive=False),
)


qgraph_option = MWOptionDecorator(
    "-g",
    "--qgraph",
    help=unwrap(
        """Location for a serialized quantum graph definition (pickle
                                              file). If this option is given then all input data options and
                                              pipeline-building options cannot be used.  Can be a URI."""
    ),
)


qgraph_id_option = MWOptionDecorator(
    "--qgraph-id",
    help=unwrap(
        """Quantum graph identifier, if specified must match the
                                                 identifier of the graph loaded from a file. Ignored if graph
                                                 is not loaded from a file."""
    ),
)


qgraph_datastore_records_option = MWOptionDecorator(
    "--qgraph-datastore-records",
    help=unwrap(
        """Include datastore records into generated quantum graph, these records are used by a
        quantum-backed butler.
        """
    ),
    is_flag=True,
)


# I wanted to use default=None here to match Python API but click silently
# replaces None with an empty tuple when multiple=True.
qgraph_node_id_option = MWOptionDecorator(
    "--qgraph-node-id",
    callback=split_commas,
    multiple=True,
    help=unwrap(
        """Only load a specified set of nodes when graph is
                                                      loaded from a file, nodes are identified by UUID
                                                      values. One or more comma-separated integers are
                                                      accepted. By default all nodes are loaded. Ignored if
                                                      graph is not loaded from a file."""
    ),
)

qgraph_header_data_option = MWOptionDecorator(
    "--show-qgraph-header",
    is_flag=True,
    default=False,
    help=unwrap(
        """Print the headerData for Quantum Graph to the
                                                           console"""
    ),
)

qgraph_dot_option = MWOptionDecorator(
    "--qgraph-dot",
    help=unwrap(
        """Location for storing GraphViz DOT representation of a
                                                  quantum graph."""
    ),
    type=MWPath(writable=True, file_okay=True, dir_okay=False),
)


replace_run_option = MWOptionDecorator(
    "--replace-run",
    help=unwrap(
        """Before creating a new RUN collection in an existing
                                                   CHAINED collection, remove the first child collection
                                                   (which must be of type RUN). This can be used to repeatedly
                                                   write to the same (parent) collection during development,
                                                   but it does not delete the datasets associated with the
                                                   replaced run unless --prune-replaced is also passed.
                                                   Requires --output, and incompatible with --extend-run."""
    ),
    is_flag=True,
)


save_pipeline_option = MWOptionDecorator(
    "-s",
    "--save-pipeline",
    help=unwrap(
        """Location for storing resulting pipeline definition in
                                                     YAML format."""
    ),
    type=MWPath(dir_okay=False, file_okay=True, writable=True),
)

save_qgraph_option = MWOptionDecorator(
    "-q",
    "--save-qgraph",
    help=unwrap(
        """URI location for storing a serialized quantum graph
                                                   definition (pickle file)."""
    ),
)


save_single_quanta_option = MWOptionDecorator(
    "--save-single-quanta",
    help=unwrap(
        """Format string of locations for storing individual
                                                          quantum graph definition (pickle files). The curly
                                                          brace {} in the input string will be replaced by a
                                                          quantum number. Can be a URI."""
    ),
)


show_option = MWOptionDecorator(
    "--show",
    callback=split_commas,
    help=unwrap(
        """Dump various info to standard output. Possible items are:
                                            `config`, `config=[Task::]<PATTERN>` or
                                            `config=[Task::]<PATTERN>:NOIGNORECASE` to dump configuration
                                            fields possibly matching given pattern and/or task label;
                                            `history=<FIELD>` to dump configuration history for a field, field
                                            name is specified as [Task::]<PATTERN>; `dump-config`,
                                            `dump-config=Task` to dump complete configuration for a task given
                                            its label or all tasks; `pipeline` to show pipeline composition;
                                            `graph` to show information about quanta; `workflow` to show
                                            information about quanta and their dependency; `tasks` to show
                                            task composition; `uri` to show predicted dataset URIs of
                                            quanta"""
    ),
    metavar="ITEM|ITEM=VALUE",
    multiple=True,
)


skip_existing_in_option = MWOptionDecorator(
    "--skip-existing-in",
    callback=split_commas,
    default=None,
    metavar="COLLECTION",
    multiple=True,
    help=unwrap(
        """If all Quantum outputs already exist in the specified list of
        collections then that Quantum will be excluded from the QuantumGraph.
        """
    ),
)


skip_existing_option = MWOptionDecorator(
    "--skip-existing",
    is_flag=True,
    help=unwrap(
        """This option is equivalent to --skip-existing-in with the name of
        the output RUN collection. If both --skip-existing-in and
        --skip-existing are given then output RUN collection is appended to
        the list of collections."""
    ),
)


clobber_outputs_option = MWOptionDecorator(
    "--clobber-outputs",
    help=(
        "Remove outputs of failed quanta from the output run when they would block the execution of new "
        "quanta with the same data ID (or assume that this will be done, if just building a QuantumGraph).  "
        "Does nothing if --extend-run is not passed."
    ),
    is_flag=True,
)


skip_init_writes_option = MWOptionDecorator(
    "--skip-init-writes",
    help=unwrap(
        """Do not write collection-wide 'init output' datasets
                                                        (e.g.schemas)."""
    ),
    is_flag=True,
)


enable_implicit_threading_option = MWOptionDecorator(
    "--enable-implicit-threading",
    help=unwrap(
        """Do not disable implicit threading use by third-party libraries (e.g. OpenBLAS).
        Implicit threading is always disabled during execution with multiprocessing."""
    ),
    is_flag=True,
)


task_option = MWOptionDecorator(
    "-t",
    "--task",
    callback=split_commas,
    help=unwrap(
        """Task name to add to pipeline, must be a fully qualified task
                                            name. Task name can be followed by colon and label name, if label
                                            is not given then task base name (class name) is used as
                                            label."""
    ),
    metavar="TASK[:LABEL]",
    multiple=True,
)


timeout_option = MWOptionDecorator(
    "--timeout", type=click.IntRange(min=0), help="Timeout for multiprocessing; maximum wall time (sec)."
)


start_method_option = MWOptionDecorator(
    "--start-method",
    default=None,
    type=click.Choice(choices=["spawn", "fork", "forkserver"]),
    help="Multiprocessing start method, default is platform-specific.",
)


fail_fast_option = MWOptionDecorator(
    "--fail-fast",
    help=unwrap(
        """Stop processing at first error, default is to process
                                                 as many tasks as possible."""
    ),
    is_flag=True,
)

save_execution_butler_option = MWOptionDecorator(
    "--save-execution-butler",
    help=unwrap(
        """Export location for an
                                                             execution-specific butler after making
                                                             QuantumGraph"""
    ),
)

mock_option = MWOptionDecorator(
    "--mock",
    help=unwrap("""Mock pipeline execution."""),
    is_flag=True,
)

unmocked_dataset_types_option = MWOptionDecorator(
    "--unmocked-dataset-types",
    callback=split_commas,
    default=None,
    metavar="COLLECTION",
    multiple=True,
    help=unwrap("""Names of input dataset types that should not be mocked."""),
)

clobber_execution_butler_option = MWOptionDecorator(
    "--clobber-execution-butler",
    help=unwrap(
        """When creating execution butler overwrite
                                                                   any existing products"""
    ),
    is_flag=True,
)

target_datastore_root_option = MWOptionDecorator(
    "--target-datastore-root",
    help=unwrap(
        """Root directory for datastore of execution butler.
        Default is to use the original datastore.
        """
    ),
)

dataset_query_constraint = MWOptionDecorator(
    "--dataset-query-constraint",
    help=unwrap(
        """When constructing a quantum graph constrain by
                                                          pre-existence of specified dataset types. Valid
                                                          values are `all` for all inputs dataset types in
                                                          pipeline, `off` to not consider dataset type
                                                          existance as a constraint, single or comma
                                                          separated list of dataset type names"""
    ),
    default="all",
)

summary_option = MWOptionDecorator(
    "--summary",
    help=(
        "Location for storing job summary (JSON file). Note that the"
        " structure of this file may not be stable."
    ),
    type=MWPath(dir_okay=False, file_okay=True, writable=True),
)


recursive_option = MWOptionDecorator(
    "--recursive",
    is_flag=True,
)

config_search_path_option = MWOptionDecorator(
    "--config-search-path",
    callback=split_commas,
    default=list(),
    help="Additional search paths for butler configuration.",
    metavar="PATH",
    multiple=True,
)

update_graph_id_option = MWOptionDecorator(
    "--update-graph-id",
    help=unwrap("Update graph ID with new unique value."),
    is_flag=True,
)

metadata_run_key_option = MWOptionDecorator(
    "--metadata-run-key",
    help=(
        "Quantum graph metadata key for the name of the output run. "
        "Empty string disables update of the metadata. "
        "Default value: output_run."
    ),
    default="output_run",
)



# This file is part of ctrl_mpexec.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

__all__ = ("SimplePipelineExecutor",)

from collections.abc import Iterable, Iterator, Mapping
from typing import Any

from lsst.daf.butler import Butler, CollectionType, Quantum
from lsst.pex.config import Config
from lsst.pipe.base import GraphBuilder, Instrument, Pipeline, PipelineTask, QuantumGraph, TaskDef

from .preExecInit import PreExecInit
from .singleQuantumExecutor import SingleQuantumExecutor
from .taskFactory import TaskFactory


class SimplePipelineExecutor:
    """A simple, high-level executor for pipelines.

    Parameters
    ----------
    quantum_graph : `~lsst.pipe.base.QuantumGraph`
        Graph to be executed.
    butler : `~lsst.daf.butler.Butler`
        Object that manages all I/O.  Must be initialized with `collections`
        and `run` properties that correspond to the input and output
        collections, which must be consistent with those used to create
        ``quantum_graph``.

    Notes
    -----
    Most callers should use one of the `classmethod` factory functions
    (`from_pipeline_filename`, `from_task_class`, `from_pipeline`) instead of
    invoking the constructor directly; these guarantee that the
    `~lsst.daf.butler.Butler` and `~lsst.pipe.base.QuantumGraph` are created
    consistently.

    This class is intended primarily to support unit testing and small-scale
    integration testing of `~lsst.pipe.base.PipelineTask` classes.  It
    deliberately lacks many features present in the command-line-only
    ``pipetask`` tool in order to keep the implementation simple.  Python
    callers that need more sophistication should call lower-level tools like
    `~lsst.pipe.base.GraphBuilder`, `PreExecInit`, and `SingleQuantumExecutor`
    directly.
    """

    def __init__(self, quantum_graph: QuantumGraph, butler: Butler):
        self.quantum_graph = quantum_graph
        self.butler = butler

    @classmethod
    def prep_butler(
        cls,
        root: str,
        inputs: Iterable[str],
        output: str,
        output_run: str | None = None,
    ) -> Butler:
        """Return configured `~lsst.daf.butler.Butler`.

        Helper method for creating `~lsst.daf.butler.Butler` instances with
        collections appropriate for processing.

        Parameters
        ----------
        root : `str`
            Root of the butler data repository; must already exist, with all
            necessary input data.
        inputs : `Iterable` [ `str` ]
            Collections to search for all input datasets, in search order.
        output : `str`
            Name of a new output `~CollectionType.CHAINED` collection to create
            that will combine both inputs and outputs.
        output_run : `str`, optional
            Name of the output `~CollectionType.RUN` that will directly hold
            all output datasets.  If not provided, a name will be created from
            ``output`` and a timestamp.

        Returns
        -------
        butler : `~lsst.daf.butler.Butler`
            Butler client instance compatible with all `classmethod` factories.
            Always writeable.
        """
        if output_run is None:
            output_run = f"{output}/{Instrument.makeCollectionTimestamp()}"
        # Make initial butler with no collections, since we haven't created
        # them yet.
        butler = Butler(root, writeable=True)
        butler.registry.registerCollection(output_run, CollectionType.RUN)
        butler.registry.registerCollection(output, CollectionType.CHAINED)
        collections = [output_run]
        collections.extend(inputs)
        butler.registry.setCollectionChain(output, collections)
        # Remake butler to let it infer default data IDs from collections, now
        # that those collections exist.
        return Butler(butler=butler, collections=[output], run=output_run)

    @classmethod
    def from_pipeline_filename(
        cls,
        pipeline_filename: str,
        *,
        where: str = "",
        bind: Mapping[str, Any] | None = None,
        butler: Butler,
    ) -> SimplePipelineExecutor:
        """Create an executor by building a QuantumGraph from an on-disk
        pipeline YAML file.

        Parameters
        ----------
        pipeline_filename : `str`
            Name of the YAML file to load the pipeline definition from.
        where : `str`, optional
            Data ID query expression that constraints the quanta generated.
        bind : `Mapping`, optional
            Mapping containing literal values that should be injected into the
            ``where`` expression, keyed by the identifiers they replace.
        butler : `~lsst.daf.butler.Butler`
            Butler that manages all I/O.  `prep_butler` can be used to create
            one.

        Returns
        -------
        executor : `SimplePipelineExecutor`
            An executor instance containing the constructed
            `~lsst.pipe.base.QuantumGraph` and `~lsst.daf.butler.Butler`,
            ready for `run` to be called.
        """
        pipeline = Pipeline.fromFile(pipeline_filename)
        return cls.from_pipeline(pipeline, butler=butler, where=where, bind=bind)

    @classmethod
    def from_task_class(
        cls,
        task_class: type[PipelineTask],
        config: Config | None = None,
        label: str | None = None,
        *,
        where: str = "",
        bind: Mapping[str, Any] | None = None,
        butler: Butler,
    ) -> SimplePipelineExecutor:
        """Create an executor by building a QuantumGraph from a pipeline
        containing a single task.

        Parameters
        ----------
        task_class : `type`
            A concrete `~lsst.pipe.base.PipelineTask` subclass.
        config : `Config`, optional
            Configuration for the task.  If not provided, task-level defaults
            will be used (no per-instrument overrides).
        label : `str`, optional
            Label for the task in its pipeline; defaults to
            ``task_class._DefaultName``.
        where : `str`, optional
            Data ID query expression that constraints the quanta generated.
        bind : `Mapping`, optional
            Mapping containing literal values that should be injected into the
            ``where`` expression, keyed by the identifiers they replace.
        butler : `~lsst.daf.butler.Butler`
            Butler that manages all I/O.  `prep_butler` can be used to create
            one.

        Returns
        -------
        executor : `SimplePipelineExecutor`
            An executor instance containing the constructed
            `~lsst.pipe.base.QuantumGraph` and `~lsst.daf.butler.Butler`,
            ready for `run` to be called.
        """
        if config is None:
            config = task_class.ConfigClass()
        if label is None:
            label = task_class._DefaultName
        if not isinstance(config, task_class.ConfigClass):
            raise TypeError(
                f"Invalid config class type: expected {task_class.ConfigClass.__name__}, "
                f"got {type(config).__name__}."
            )
        task_def = TaskDef(taskName=task_class.__name__, config=config, label=label, taskClass=task_class)
        return cls.from_pipeline([task_def], butler=butler, where=where, bind=bind)

    @classmethod
    def from_pipeline(
        cls,
        pipeline: Pipeline | Iterable[TaskDef],
        *,
        where: str = "",
        bind: Mapping[str, Any] | None = None,
        butler: Butler,
        **kwargs: Any,
    ) -> SimplePipelineExecutor:
        """Create an executor by building a QuantumGraph from an in-memory
        pipeline.

        Parameters
        ----------
        pipeline : `Pipeline` or `Iterable` [ `TaskDef` ]
            A Python object describing the tasks to run, along with their
            labels and configuration.
        where : `str`, optional
            Data ID query expression that constraints the quanta generated.
        bind : `Mapping`, optional
            Mapping containing literal values that should be injected into the
            ``where`` expression, keyed by the identifiers they replace.
        butler : `~lsst.daf.butler.Butler`
            Butler that manages all I/O.  `prep_butler` can be used to create
            one.

        Returns
        -------
        executor : `SimplePipelineExecutor`
            An executor instance containing the constructed
            `~lsst.pipe.base.QuantumGraph` and `~lsst.daf.butler.Butler`,
            ready for `run` to be called.
        """
        if isinstance(pipeline, Pipeline):
            pipeline = list(pipeline.toExpandedPipeline())
        else:
            pipeline = list(pipeline)
        graph_builder = GraphBuilder(butler.registry)
        assert butler.run is not None, "Butler output run collection must be defined"
        quantum_graph = graph_builder.makeGraph(
            pipeline, collections=butler.collections, run=butler.run, userQuery=where, bind=bind
        )
        return cls(quantum_graph=quantum_graph, butler=butler)

    def run(self, register_dataset_types: bool = False, save_versions: bool = True) -> list[Quantum]:
        """Run all the quanta in the `~lsst.pipe.base.QuantumGraph` in
        topological order.

        Use this method to run all quanta in the graph.  Use
        `as_generator` to get a generator to run the quanta one at
        a time.

        Parameters
        ----------
        register_dataset_types : `bool`, optional
            If `True`, register all output dataset types before executing any
            quanta.
        save_versions : `bool`, optional
            If `True` (default), save a package versions dataset.

        Returns
        -------
        quanta : `list` [ `~lsst.daf.butler.Quantum` ]
            Executed quanta.

        Notes
        -----
        A topological ordering is not in general unique, but no other
        guarantees are made about the order in which quanta are processed.
        """
        return list(
            self.as_generator(register_dataset_types=register_dataset_types, save_versions=save_versions)
        )

    def as_generator(
        self, register_dataset_types: bool = False, save_versions: bool = True
    ) -> Iterator[Quantum]:
        """Yield quanta in the `~lsst.pipe.base.QuantumGraph` in topological
        order.

        These quanta will be run as the returned generator is iterated
        over.  Use this method to run the quanta one at a time.
        Use `run` to run all quanta in the graph.

        Parameters
        ----------
        register_dataset_types : `bool`, optional
            If `True`, register all output dataset types before executing any
            quanta.
        save_versions : `bool`, optional
            If `True` (default), save a package versions dataset.

        Returns
        -------
        quanta : `~collections.abc.Iterator` [ `~lsst.daf.butler.Quantum` ]
            Executed quanta.

        Notes
        -----
        Global initialization steps (see `PreExecInit`) are performed
        immediately when this method is called, but individual quanta are not
        actually executed until the returned iterator is iterated over.

        A topological ordering is not in general unique, but no other
        guarantees are made about the order in which quanta are processed.
        """
        task_factory = TaskFactory()
        pre_exec_init = PreExecInit(self.butler, task_factory)
        pre_exec_init.initialize(
            graph=self.quantum_graph, registerDatasetTypes=register_dataset_types, saveVersions=save_versions
        )
        single_quantum_executor = SingleQuantumExecutor(self.butler, task_factory)
        # Important that this returns a generator expression rather than being
        # a generator itself; that is what makes the PreExecInit stuff above
        # happen immediately instead of when the first quanta is executed,
        # which might be useful for callers who want to check the state of the
        # repo in between.
        return (single_quantum_executor.execute(qnode.taskDef, qnode.quantum) for qnode in self.quantum_graph)



# This file is part of ctrl_mpexec.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os
import shutil
import tempfile
import unittest
from typing import Any, Dict

import lsst.daf.butler
import lsst.utils.tests
from lsst.ctrl.mpexec import SimplePipelineExecutor
from lsst.pex.config import Field
from lsst.pipe.base import (
    PipelineTask,
    PipelineTaskConfig,
    PipelineTaskConnections,
    Struct,
    TaskDef,
    TaskMetadata,
    connectionTypes,
)
from lsst.pipe.base.tests.no_dimensions import NoDimensionsTestTask
from lsst.utils.introspection import get_full_type_name

TESTDIR = os.path.abspath(os.path.dirname(__file__))


class NoDimensionsTestConnections2(PipelineTaskConnections, dimensions=set()):
    input = connectionTypes.Input(
        name="input", doc="some dict-y input data for testing", storageClass="TaskMetadataLike"
    )
    output = connectionTypes.Output(
        name="output", doc="some dict-y output data for testing", storageClass="StructuredDataDict"
    )


class NoDimensionsTestConfig2(PipelineTaskConfig, pipelineConnections=NoDimensionsTestConnections2):
    key = Field(dtype=str, doc="String key for the dict entry the task sets.", default="one")
    value = Field(dtype=int, doc="Integer value for the dict entry the task sets.", default=1)
    outputSC = Field(dtype=str, doc="Output storage class requested", default="dict")


class NoDimensionsMetadataTestConnections(PipelineTaskConnections, dimensions=set()):
    input = connectionTypes.Input(
        name="input", doc="some dict-y input data for testing", storageClass="StructuredDataDict"
    )
    # Deliberately choose a storage class that does not match the metadata
    # default TaskMetadata storage class.
    meta = connectionTypes.Input(
        name="a_metadata", doc="Metadata from previous task", storageClass="StructuredDataDict"
    )
    output = connectionTypes.Output(
        name="output", doc="some dict-y output data for testing", storageClass="StructuredDataDict"
    )


class NoDimensionsMetadataTestConfig(
    PipelineTaskConfig, pipelineConnections=NoDimensionsMetadataTestConnections
):
    key = Field(dtype=str, doc="String key for the dict entry the task sets.", default="one")
    value = Field(dtype=int, doc="Integer value for the dict entry the task sets.", default=1)
    outputSC = Field(dtype=str, doc="Output storage class requested", default="dict")


class NoDimensionsMetadataTestTask(PipelineTask):
    """A simple pipeline task that can take a metadata as input."""

    ConfigClass = NoDimensionsMetadataTestConfig
    _DefaultName = "noDimensionsMetadataTest"

    def run(self, input: Dict[str, int], meta: Dict[str, Any]) -> Struct:
        """Run the task, adding the configured key-value pair to the input
        argument and returning it as the output.

        Parameters
        ----------
        input : `dict`
            Dictionary to update and return.

        Returns
        -------
        result : `lsst.pipe.base.Struct`
            Struct with a single ``output`` attribute.
        """
        self.log.info("Run metadata method given data of type: %s", get_full_type_name(input))
        output = input.copy()
        output[self.config.key] = self.config.value

        self.log.info("Received task metadata (%s): %s", get_full_type_name(meta), meta)

        # Can change the return type via configuration.
        if "TaskMetadata" in self.config.outputSC:
            output = TaskMetadata.from_dict(output)
        elif type(output) == TaskMetadata:
            # Want the output to be a dict
            output = output.to_dict()
        self.log.info("Run method returns data of type: %s", get_full_type_name(output))
        return Struct(output=output)


class SimplePipelineExecutorTests(lsst.utils.tests.TestCase):
    """Test the SimplePipelineExecutor API with a trivial task."""

    def setUp(self):
        self.path = tempfile.mkdtemp()
        # standalone parameter forces the returned config to also include
        # the information from the search paths.
        config = lsst.daf.butler.Butler.makeRepo(
            self.path, standalone=True, searchPaths=[os.path.join(TESTDIR, "config")]
        )
        self.butler = SimplePipelineExecutor.prep_butler(config, [], "fake")
        self.butler.registry.registerDatasetType(
            lsst.daf.butler.DatasetType(
                "input",
                dimensions=self.butler.dimensions.empty,
                storageClass="StructuredDataDict",
            )
        )
        self.butler.put({"zero": 0}, "input")

    def tearDown(self):
        shutil.rmtree(self.path, ignore_errors=True)

    def test_from_task_class(self):
        """Test executing a single quantum with an executor created by the
        `from_task_class` factory method, and the
        `SimplePipelineExecutor.as_generator` method.
        """
        executor = SimplePipelineExecutor.from_task_class(NoDimensionsTestTask, butler=self.butler)
        (quantum,) = executor.as_generator(register_dataset_types=True)
        self.assertEqual(self.butler.get("output"), {"zero": 0, "one": 1})

    def _configure_pipeline(self, config_a_cls, config_b_cls, storageClass_a=None, storageClass_b=None):
        """Configure a pipeline with from_pipeline."""

        config_a = config_a_cls()
        config_a.connections.output = "intermediate"
        if storageClass_a:
            config_a.outputSC = storageClass_a
        config_b = config_b_cls()
        config_b.connections.input = "intermediate"
        if storageClass_b:
            config_b.outputSC = storageClass_b
        config_b.key = "two"
        config_b.value = 2
        task_defs = [
            TaskDef(label="a", taskClass=NoDimensionsTestTask, config=config_a),
            TaskDef(label="b", taskClass=NoDimensionsTestTask, config=config_b),
        ]
        executor = SimplePipelineExecutor.from_pipeline(task_defs, butler=self.butler)
        return executor

    def _test_logs(self, log_output, input_type_a, output_type_a, input_type_b, output_type_b):
        """Check the expected input types received by tasks A and B"""
        all_logs = "\n".join(log_output)
        self.assertIn(f"lsst.a:Run method given data of type: {input_type_a}", all_logs)
        self.assertIn(f"lsst.b:Run method given data of type: {input_type_b}", all_logs)
        self.assertIn(f"lsst.a:Run method returns data of type: {output_type_a}", all_logs)
        self.assertIn(f"lsst.b:Run method returns data of type: {output_type_b}", all_logs)

    def test_from_pipeline(self):
        """Test executing a two quanta from different configurations of the
        same task, with an executor created by the `from_pipeline` factory
        method, and the `SimplePipelineExecutor.run` method.
        """
        executor = self._configure_pipeline(
            NoDimensionsTestTask.ConfigClass, NoDimensionsTestTask.ConfigClass
        )

        with self.assertLogs("lsst", level="INFO") as cm:
            quanta = executor.run(register_dataset_types=True, save_versions=False)
        self._test_logs(cm.output, "dict", "dict", "dict", "dict")

        self.assertEqual(len(quanta), 2)
        self.assertEqual(self.butler.get("intermediate"), {"zero": 0, "one": 1})
        self.assertEqual(self.butler.get("output"), {"zero": 0, "one": 1, "two": 2})

    def test_from_pipeline_intermediates_differ(self):
        """Run pipeline but intermediates definition in registry differs."""
        executor = self._configure_pipeline(
            NoDimensionsTestTask.ConfigClass,
            NoDimensionsTestTask.ConfigClass,
            storageClass_b="TaskMetadataLike",
        )

        # Pre-define the "intermediate" storage class to be something that is
        # like a dict but is not a dict. This will fail unless storage
        # class conversion is supported in put and get.
        self.butler.registry.registerDatasetType(
            lsst.daf.butler.DatasetType(
                "intermediate",
                dimensions=self.butler.dimensions.empty,
                storageClass="TaskMetadataLike",
            )
        )

        with self.assertLogs("lsst", level="INFO") as cm:
            quanta = executor.run(register_dataset_types=True, save_versions=False)
        # A dict is given to task a without change.
        # A returns a dict because it has not been told to do anything else.
        # That does not match the storage class so it will be converted
        # on put.
        # b is given a dict, because that's what its connection asks for.
        # b returns a TaskMetadata because that's how we configured it, but
        # the butler expects a dict so it is converted on put.
        self._test_logs(cm.output, "dict", "dict", "dict", "lsst.pipe.base.TaskMetadata")

        self.assertEqual(len(quanta), 2)
        self.assertEqual(self.butler.get("intermediate").to_dict(), {"zero": 0, "one": 1})
        self.assertEqual(self.butler.get("output"), {"zero": 0, "one": 1, "two": 2})

    def test_from_pipeline_output_differ(self):
        """Run pipeline but output definition in registry differs."""
        executor = self._configure_pipeline(
            NoDimensionsTestTask.ConfigClass,
            NoDimensionsTestTask.ConfigClass,
            storageClass_a="TaskMetadataLike",
        )

        # Pre-define the "output" storage class to be something that is
        # like a dict but is not a dict. This will fail unless storage
        # class conversion is supported in put and get.
        self.butler.registry.registerDatasetType(
            lsst.daf.butler.DatasetType(
                "output",
                dimensions=self.butler.dimensions.empty,
                storageClass="TaskMetadataLike",
            )
        )

        with self.assertLogs("lsst", level="INFO") as cm:
            quanta = executor.run(register_dataset_types=True, save_versions=False)
        # a has been told to return a TaskMetadata but will convert to dict.
        # b returns a dict and that is converted to TaskMetadata on put.
        self._test_logs(cm.output, "dict", "lsst.pipe.base.TaskMetadata", "dict", "dict")

        self.assertEqual(len(quanta), 2)
        self.assertEqual(self.butler.get("intermediate"), {"zero": 0, "one": 1})
        self.assertEqual(self.butler.get("output").to_dict(), {"zero": 0, "one": 1, "two": 2})

    def test_from_pipeline_input_differ(self):
        """Run pipeline but input definition in registry differs."""

        # This config declares that the pipeline takes a TaskMetadata
        # as input but registry already thinks it has a StructureDataDict.
        executor = self._configure_pipeline(NoDimensionsTestConfig2, NoDimensionsTestTask.ConfigClass)

        with self.assertLogs("lsst", level="INFO") as cm:
            quanta = executor.run(register_dataset_types=True, save_versions=False)
        self._test_logs(cm.output, "lsst.pipe.base.TaskMetadata", "dict", "dict", "dict")

        self.assertEqual(len(quanta), 2)
        self.assertEqual(self.butler.get("intermediate"), {"zero": 0, "one": 1})
        self.assertEqual(self.butler.get("output"), {"zero": 0, "one": 1, "two": 2})

    def test_from_pipeline_incompatible(self):
        """Run pipeline but definitions are not compatible."""
        executor = self._configure_pipeline(
            NoDimensionsTestTask.ConfigClass, NoDimensionsTestTask.ConfigClass
        )

        # Incompatible output dataset type.
        self.butler.registry.registerDatasetType(
            lsst.daf.butler.DatasetType(
                "output",
                dimensions=self.butler.dimensions.empty,
                storageClass="StructuredDataList",
            )
        )

        with self.assertRaisesRegex(
            ValueError, "StructuredDataDict.*inconsistent with registry definition.*StructuredDataList"
        ):
            executor.run(register_dataset_types=True, save_versions=False)

    def test_from_pipeline_metadata(self):
        """Test two tasks where the output uses metadata from input."""
        # Must configure a special pipeline for this test.
        config_a = NoDimensionsTestTask.ConfigClass()
        config_a.connections.output = "intermediate"
        config_b = NoDimensionsMetadataTestTask.ConfigClass()
        config_b.connections.input = "intermediate"
        config_b.key = "two"
        config_b.value = 2
        task_defs = [
            TaskDef(label="a", taskClass=NoDimensionsTestTask, config=config_a),
            TaskDef(label="b", taskClass=NoDimensionsMetadataTestTask, config=config_b),
        ]
        executor = SimplePipelineExecutor.from_pipeline(task_defs, butler=self.butler)

        with self.assertLogs("test_simple_pipeline_executor", level="INFO") as cm:
            quanta = executor.run(register_dataset_types=True, save_versions=False)
        self.assertIn(f"Received task metadata ({get_full_type_name(dict)})", "".join(cm.output))

        self.assertEqual(len(quanta), 2)
        self.assertEqual(self.butler.get("intermediate"), {"zero": 0, "one": 1})
        self.assertEqual(self.butler.get("output"), {"zero": 0, "one": 1, "two": 2})

    def test_from_pipeline_file(self):
        """Test executing a two quanta from different configurations of the
        same task, with an executor created by the `from_pipeline_filename`
        factory method, and the `SimplePipelineExecutor.run` method.
        """
        filename = os.path.join(self.path, "pipeline.yaml")
        with open(filename, "w") as f:
            f.write(
                """
                description: test
                tasks:
                    a:
                        class: "lsst.pipe.base.tests.no_dimensions.NoDimensionsTestTask"
                        config:
                            connections.output: "intermediate"
                    b:
                        class: "lsst.pipe.base.tests.no_dimensions.NoDimensionsTestTask"
                        config:
                            connections.input: "intermediate"
                            key: "two"
                            value: 2
                """
            )
        executor = SimplePipelineExecutor.from_pipeline_filename(filename, butler=self.butler)
        quanta = executor.run(register_dataset_types=True, save_versions=False)
        self.assertEqual(len(quanta), 2)
        self.assertEqual(self.butler.get("intermediate"), {"zero": 0, "one": 1})
        self.assertEqual(self.butler.get("output"), {"zero": 0, "one": 1, "two": 2})


class MemoryTester(lsst.utils.tests.MemoryTestCase):
    pass


def setup_module(module):
    lsst.utils.tests.init()


if __name__ == "__main__":
    lsst.utils.tests.init()
    unittest.main()
